# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
#
# Copyright 2008-2021 Neongecko.com Inc. | All Rights Reserved
#
# Notice of License - Duplicating this Notice of License near the start of any file containing
# a derivative of this software is a condition of license for this software.
# Friendly Licensing:
# No charge, open source royalty free use of the Neon AI software source and object is offered for
# educational users, noncommercial enthusiasts, Public Benefit Corporations (and LLCs) and
# Social Purpose Corporations (and LLCs). Developers can contact developers@neon.ai
# For commercial licensing, distribution of derivative works or redistribution please contact licenses@neon.ai
# Distributed on an "AS IS” basis without warranties or conditions of any kind, either express or implied.
# Trademarks of Neongecko: Neon AI(TM), Neon Assist (TM), Neon Communicator(TM), Klat(TM)
# Authors: Guy Daniels, Daniel McKnight, Regina Bloomstine, Elon Gasper, Richard Leeds
#
# Specialized conversational reconveyance options from Conversation Processing Intelligence Corp.
# US Patents 2008-2021: US7424516, US20140161250, US20140177813, US8638908, US8068604, US8553852, US10530923, US10530924
# China Patent: CN102017585  -  Europe Patent: EU2156652  -  Patents Pending

import os
from datetime import timedelta, datetime
from typing import Optional

from ovos_utils.signal import ensure_directory_exists

from neon_utils.logger import LOG


def get_likes_from_csv(file: str, user: Optional[str],
                       line_limit: int = 500, date_limit: Optional[timedelta] = timedelta(days=7)) -> dict:
    """
    Reads likes data from the specified csv file (default: selected_ts.csv). The lesser of the passed limits will be
    evaluated. Pass `None` for user to get data for all users within specified limits
    :param file: path to csv file to evaluate
    :param user: username associated with likes (None for all)
    :param line_limit: maximum number of lines to process
    :param date_limit: maximum time history to process
    :return: dict of likes data
    """
    file = os.path.expanduser(file)
    if not os.path.isfile(file):
        raise FileNotFoundError
    with open(file, "r") as csv:
        lines_to_evaluate = csv.readlines()

    if line_limit and len(lines_to_evaluate) > line_limit:
        lines_to_evaluate = lines_to_evaluate[-line_limit:]
    else:
        lines_to_evaluate = lines_to_evaluate[1:]
    lines_to_evaluate.reverse()
    likes = {}
    for line in lines_to_evaluate:
        date_str, profile, device, phrase, transcription_filename, brand = line.rstrip("\n").split(",", 5)
        # Date,Profile,Device,Phrase,Instance,Brand
        if date_limit and datetime.now() - datetime.strptime(date_str, "%Y-%m-%d") > date_limit:
            # Stop parsing once we reach our oldest date limit
            break
        if user and profile != user:
            # Ignore anything from other users than requested
            continue

        transcription_subdir = f"{profile}-{date_str}"
        if brand not in likes.keys():
            likes[brand] = []
        likes[brand].append({"date": date_str,
                             "user": profile,
                             "file": os.path.join(transcription_subdir, transcription_filename),
                             "utterance": phrase})
    return likes


def update_csv(info: list, location: str):
    """
    Neon function used to generate CSV files for transcripts
    :param info: list of data to export to csv
    :param location: fully defined file path
    :return:
    """
    import csv

    def _write_line(data, file):
        try:
            with open(file, 'a+') as to_write:
                writer = csv.writer(to_write)
                writer.writerow(data)
        except IOError as e:
            LOG.error(e)

    location = os.path.expanduser(location)

    # Check if file exists with first row
    csv_dir = os.path.dirname(location)
    if not os.path.isfile(location):
        os.makedirs(csv_dir, exist_ok=True)
        if os.path.basename(location) == "full_ts.csv":
            list_transcription_headers = ["Date", "Time", "Profile", "Device", "Input", "Location", "Wav_Length"]
            _write_line(list_transcription_headers, location)
        elif os.path.basename(location) == "selected_ts.csv":
            list_selected_headers = ["Date", "Profile", "Device", "Phrase", "Instance", "Brand"]
            _write_line(list_selected_headers, location)
        elif os.path.basename(location) == "running_out.csv":
            list_selected_headers = ["Date", "Profile", "Device", "Phrase", "Instance", "Item"]
            _write_line(list_selected_headers, location)
        elif os.path.basename(location) == "symptom_checker.csv":
            list_selected_headers = ["Date", "Profile", "Agent", "Device", "Phrase", "Tags", "Location"]
            _write_line(list_selected_headers, location)
    # Write out  data
    _write_line(info, location)


def write_transcript_file(utterance: str, path: str, transcript_name: str, username: str, timestamp: float) -> str:
    """
    Writes out a transcript at the specified path and filename
    :param utterance: Text to transcribe
    :param path: Path to transcriptions
    :param transcript_name: Name of transcript (directory name)
    :param username: User associated with utterance
    :param timestamp: Timestamp associated with utterance
    :return: formatted line written to the transcript
    """
    import datetime

    path = os.path.expanduser(path)
    ensure_directory_exists(path, transcript_name)

    f_time = str(datetime.datetime.fromtimestamp(timestamp))
    f_date = str(datetime.date.fromtimestamp(timestamp))
    shared_file = os.path.join(path, transcript_name, f"{f_date}.txt")
    person_file = os.path.join(path, transcript_name, f"{username}-{f_date}.txt")

    formatted_line = f"{username}-{f_time} {utterance}\n"

    with open(shared_file, 'a+') as shared:
        shared.write(formatted_line)
    with open(person_file, 'a+') as person:
        person.write(formatted_line)
    return formatted_line


def get_transcript_file(path: str, transcript_name: str, username: Optional[str], f_date: str) -> Optional[str]:
    """
    Returns the path to a requested transcript file if exists
    :param path: Path to transcripts
    :param transcript_name: Name of transcript to return
        (ts_selected_transcripts, running_out_transcripts, ts_transcripts, etc.)
    :param username: Username requested (None for combined daily transcripts)
    :param f_date: String formatted date requested (YYYY-MM-DD)
    :return: Path to requested transcript (None if not exists)
    """
    path = os.path.expanduser(path)
    if username:
        transcript_file = os.path.join(path, transcript_name, f"{username}-{f_date}.txt")
    else:
        transcript_file = os.path.join(path, transcript_name, f"{f_date}.txt")

    if os.path.isfile(transcript_file):
        return transcript_file
    else:
        return None



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
#
# Copyright 2008-2021 Neongecko.com Inc. | All Rights Reserved
#
# Notice of License - Duplicating this Notice of License near the start of any file containing
# a derivative of this software is a condition of license for this software.
# Friendly Licensing:
# No charge, open source royalty free use of the Neon AI software source and object is offered for
# educational users, noncommercial enthusiasts, Public Benefit Corporations (and LLCs) and
# Social Purpose Corporations (and LLCs). Developers can contact developers@neon.ai
# For commercial licensing, distribution of derivative works or redistribution please contact licenses@neon.ai
# Distributed on an "AS IS” basis without warranties or conditions of any kind, either express or implied.
# Trademarks of Neongecko: Neon AI(TM), Neon Assist (TM), Neon Communicator(TM), Klat(TM)
# Authors: Guy Daniels, Daniel McKnight, Regina Bloomstine, Elon Gasper, Richard Leeds
#
# Specialized conversational reconveyance options from Conversation Processing Intelligence Corp.
# US Patents 2008-2021: US7424516, US20140161250, US20140177813, US8638908, US8068604, US8553852, US10530923, US10530924
# China Patent: CN102017585  -  Europe Patent: EU2156652  -  Patents Pending

import json
import os.path
from typing import Optional
from urllib.parse import urlparse
from neon_utils.logger import LOG


def find_neon_git_token(base_path: str = "~/") -> str:
    """
    Searches environment variables and standard locations for a text file with a Github token.
    Args:
        base_path: Base directory to check in addition to XDG directories (default ~/)
    Returns:
        Github token string
    """
    path_to_check = os.path.expanduser(base_path)
    paths_to_check = (path_to_check,
                      os.path.join(path_to_check, "token.txt"),
                      os.path.expanduser("~/.local/share/neon/git_token.txt"),
                      os.path.expanduser("~/.local/share/neon/token.txt"))
    cred_file = None
    for path in paths_to_check:
        if os.path.isfile(path):
            cred_file = path
            break
    if not cred_file:
        if os.environ.get("GITHUB_TOKEN"):
            return os.environ.get("GITHUB_TOKEN")
        raise FileNotFoundError("Could not locate github credentials")

    with open(cred_file) as f:
        git_token = f.read().strip('\n').strip()
    if not git_token:
        raise ValueError(f"Empty Git token specified in {cred_file}")
    return git_token


def find_neon_aws_keys(base_path: str = "~/") -> dict:
    """
    Searches standard locations for AWS credentials
    Args:
        base_path: Base directory to check in addition to XDG directories (default ~/)
    Returns:
        dict containing 'aws_access_key_id' and 'aws_secret_access_key'
    """
    path_to_check = os.path.expanduser(base_path)
    csv_paths = (path_to_check,
                 os.path.join(path_to_check, "accessKeys.csv"))
    sys_path = os.path.expanduser("~/.aws/credentials")
    json_path = os.path.expanduser("~/.local/share/neon/aws.json")

    amazon_creds = None
    for path in csv_paths:
        if os.path.isfile(path):
            try:
                with open(path, "r") as f:
                    aws_id, aws_key = f.readlines()[1].rstrip('\n').split(',', 1)
                    amazon_creds = {"aws_access_key_id": aws_id,
                                    "aws_secret_access_key": aws_key}

            except Exception as e:
                LOG.error(e)
                LOG.error(path)

    if not amazon_creds:
        if os.path.isfile(json_path):
            with open(json_path, "r") as f:
                amazon_creds = json.load(f)
        elif os.path.isfile(sys_path):
            with open(sys_path, "r") as f:
                for line in f.read().split("\n"):
                    if line.startswith("aws_access_key_id"):
                        aws_id = line.split("=", 1)[1].strip()
                    elif line.startswith("aws_secret_access_key"):
                        aws_key = line.split("=", 1)[1].strip()
            amazon_creds = {"aws_access_key_id": aws_id,
                            "aws_secret_access_key": aws_key}

    if not amazon_creds:
        raise FileNotFoundError(f"No aws credentials found in default locations or path: {path_to_check}")
    return amazon_creds


def find_neon_google_keys(base_path: str = "~/") -> dict:
    """
    Locates google json credentials and returns the parsed credentials as a dict
    Args:
        base_path: Base directory to check in addition to XDG directories (default ~/)
    Returns:
        dict Google json credential
    """
    path_to_check = os.path.expanduser(base_path)
    paths_to_check = (path_to_check,
                      os.path.join(path_to_check, "google.json"),
                      os.path.expanduser("~/.local/share/neon/google.json"))
    for path in paths_to_check:
        if os.path.isfile(path):
            try:
                with open(path, "r") as f:
                    credential = json.load(f)
                return credential
            except Exception as e:
                LOG.error(f"Invalid google credential found at: {path}")
                raise e
    raise FileNotFoundError(f"No google credentials found in default locations or path: {path_to_check}")


def populate_amazon_keys_config(aws_keys: dict, config_path: Optional[str] = None):
    """
    Populates configuration with the specified Amazon keys to be referenced by tts/translation modules.
    Args:
        aws_keys: Dict of aws credentials to use (returned by `find_neon_aws_keys()`)
        config_path: Override path to ngi_local_conf
    """
    from neon_utils.configuration_utils import NGIConfig

    assert "aws_access_key_id" in aws_keys
    assert "aws_secret_access_key" in aws_keys

    if not aws_keys.get("aws_access_key_id") or not aws_keys.get("aws_secret_access_key"):
        raise ValueError

    local_conf = NGIConfig("ngi_local_conf", config_path, True)
    aws_config = local_conf["tts"]["amazon"]
    aws_config = {**aws_config, **aws_keys}
    local_conf["tts"]["amazon"] = aws_config
    local_conf.write_changes()
    # TODO: This should move to auth config when available DM


def populate_github_token_config(token: str, config_path: Optional[str] = None):
    """
    Populates configuration with the specified github token for later reference.
    Args:
        token: String Github token
        config_path: Override path to ngi_local_conf
    """
    from neon_utils.configuration_utils import NGIConfig

    assert token

    local_conf = NGIConfig("ngi_local_conf", config_path, True)
    local_conf["skills"]["neon_token"] = token
    local_conf.write_changes()
    # TODO: This should move to auth config when available DM


def repo_is_neon(repo_url: str) -> bool:
    """
    Determines if the specified repository url is part of the NeonGeckoCom org on github
    Args:
        repo_url: string url to check
    Returns:
        True if the repository URL is known to be accessible using a neon auth key
    """
    url = urlparse(repo_url)
    if not url.scheme or not url.netloc:
        raise ValueError(f"{repo_url} is not a valid url")
    if any([x for x in ("github.com", "githubusercontent.com") if x in url.netloc]):
        try:
            author = url.path.split('/')[1]
        except IndexError:
            raise ValueError(f"{repo_url} is not a valid github url")
        if author.lower() == "neongeckocom":
            return True
        elif author.lower().startswith("neon"):  # TODO: Get token and scrape org? DM
            LOG.info(f"Assuming repository uses Neon auth: {repo_url}")
            return True
    return False



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
#
# Copyright 2008-2021 Neongecko.com Inc. | All Rights Reserved
#
# Notice of License - Duplicating this Notice of License near the start of any file containing
# a derivative of this software is a condition of license for this software.
# Friendly Licensing:
# No charge, open source royalty free use of the Neon AI software source and object is offered for
# educational users, noncommercial enthusiasts, Public Benefit Corporations (and LLCs) and
# Social Purpose Corporations (and LLCs). Developers can contact developers@neon.ai
# For commercial licensing, distribution of derivative works or redistribution please contact licenses@neon.ai
# Distributed on an "AS IS” basis without warranties or conditions of any kind, either express or implied.
# Trademarks of Neongecko: Neon AI(TM), Neon Assist (TM), Neon Communicator(TM), Klat(TM)
# Authors: Guy Daniels, Daniel McKnight, Regina Bloomstine, Elon Gasper, Richard Leeds
#
# Specialized conversational reconveyance options from Conversation Processing Intelligence Corp.
# US Patents 2008-2021: US7424516, US20140161250, US20140177813, US8638908, US8068604, US8553852, US10530923, US10530924
# China Patent: CN102017585  -  Europe Patent: EU2156652  -  Patents Pending

import setuptools

with open("README.md", "r") as f:
    long_description = f.read()

with open("./version.py", "r", encoding="utf-8") as v:
    for line in v.readlines():
        if line.startswith("__version__"):
            if '"' in line:
                version = line.split('"')[1]
            else:
                version = line.split("'")[1]

with open("./requirements.txt", "r", encoding="utf-8") as r:
    requirements = r.readlines()

setuptools.setup(
    name="neon-utils",
    version=version,
    author="NeonDaniel",
    author_email="daniel@neon.ai",
    description="Utilities for running Neon skills in Mycroft/OVOS",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/neongeckocom/neon-skill-utils",
    packages=setuptools.find_packages(),
    package_data={'': ['*.yml']},
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent"
    ],
    python_requires='>=3.6',
    install_requires=requirements,
    entry_points={
        'console_scripts': [
            "neon-config-import=neon_utils.configuration_utils:create_config_from_setup_params"
        ]
    }
)



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
#
# Copyright 2008-2021 Neongecko.com Inc. | All Rights Reserved
#
# Notice of License - Duplicating this Notice of License near the start of any file containing
# a derivative of this software is a condition of license for this software.
# Friendly Licensing:
# No charge, open source royalty free use of the Neon AI software source and object is offered for
# educational users, noncommercial enthusiasts, Public Benefit Corporations (and LLCs) and
# Social Purpose Corporations (and LLCs). Developers can contact developers@neon.ai
# For commercial licensing, distribution of derivative works or redistribution please contact licenses@neon.ai
# Distributed on an "AS IS” basis without warranties or conditions of any kind, either express or implied.
# Trademarks of Neongecko: Neon AI(TM), Neon Assist (TM), Neon Communicator(TM), Klat(TM)
# Authors: Guy Daniels, Daniel McKnight, Regina Bloomstine, Elon Gasper, Richard Leeds
#
# Specialized conversational reconveyance options from Conversation Processing Intelligence Corp.
# US Patents 2008-2021: US7424516, US20140161250, US20140177813, US8638908, US8068604, US8553852, US10530923, US10530924
# China Patent: CN102017585  -  Europe Patent: EU2156652  -  Patents Pending

from neon_utils.logger import LOG


def search_convo_dict(id_to_keywords: dict, search_term: str,
                      max_matches_to_return: int = 5, handle_typos: bool = True) -> list:
    """
    Searches the passed id_to_keywords dict for id's matching the search_term
    :param id_to_keywords: dict of id to list of keywords to match
    :param search_term: string search term to locate
    :param max_matches_to_return: max number of sorted results to return
    :param handle_typos: boolean to allow inexact string matches
    :return: list of matched IDs
    """
    def _return_close_values(source_dict, to_find, num_of_return, typo_check=True) -> list:
        import difflib
        allow_diffs = 1  # if len(to_find) < 4 else 2

        perfect_match = [i for i, v in source_dict.items() if (isinstance(v, str) and to_find == v)
                         or (isinstance(v, list) and [x for x in v if to_find == x])]
        contains_in = [k for k, i in source_dict.items() if to_find in i or (isinstance(i, str) and i in to_find)
                       or (isinstance(i, list) and [x for x in i if x in to_find or to_find in x])]
        close_match = list(set(k for c in difflib.get_close_matches(to_find, source_dict.values())
                               for k, v in source_dict.items() if v == c))

        def check_diffs(i, k):
            len_i = len(i)
            len_tf = len(to_find)
            if abs(len_i - len_tf) > allow_diffs:
                return
            counter_i, counter_tf = 0, 0
            diff_count = 0
            while counter_i < len_i and counter_tf < len_tf:
                if to_find[counter_tf] != i[counter_i]:
                    diff_count += 1
                    if diff_count > allow_diffs:
                        break
                    if len_tf > len_i:
                        counter_tf += 1
                    elif len_i > len_tf:
                        counter_i += 1
                    else:
                        counter_i += 1
                        counter_tf += 1
                else:
                    counter_i += 1
                    counter_tf += 1

            if diff_count <= allow_diffs:
                typos_check.append(k)

        typos_check = []
        if typo_check:
            for k, i in source_dict.items():
                if isinstance(i, str):
                    check_diffs(i, k)
                elif isinstance(i, list):
                    for x in i:
                        if k not in typos_check:
                            check_diffs(x, k)

        LOG.debug(f"perfect match - {perfect_match}")
        LOG.debug(f'contains part - {contains_in}')
        LOG.debug(f"close_match - {close_match}")
        if typo_check:
            LOG.debug(f"typo check {allow_diffs} characters diff - {typos_check}")

        if not close_match and not contains_in and not perfect_match and not typos_check:
            return []
        try:
            res_tmp = typos_check
            res_tmp.extend([i for i in close_match if i not in typos_check])
            res_tmp.extend([i for i in contains_in if i not in close_match])
            LOG.debug(f'res_tmp is {res_tmp}')
        except Exception as x:
            LOG.debug(x)
            res_tmp = []

        if perfect_match:
            if len(perfect_match) == num_of_return:
                return perfect_match
            else:
                need_add = num_of_return - len(perfect_match)
                LOG.debug(len(perfect_match))
                LOG.debug(need_add)
                if need_add < 0:
                    return perfect_match[:num_of_return]

                res = perfect_match
                if res_tmp:
                    res.extend([i for i in res_tmp if i not in perfect_match])
                    LOG.debug(res)
        else:
            res = res_tmp

        if not res:
            raise ValueError("No results matched!")
        return res[:num_of_return]

    fin_res = []
    from collections import Counter
    import operator
    if not search_term:
        return []
    if " " in search_term:
        for i in search_term.split(" "):
            fin_res.extend(_return_close_values(id_to_keywords, i, max_matches_to_return, handle_typos))
        LOG.debug(fin_res)
        tmp_f = Counter(fin_res)
        LOG.debug(tmp_f)
        tmp = sorted(tmp_f.items(), key=operator.itemgetter(1), reverse=True)
        LOG.debug(tmp)
        return [i for i, m in tmp][:max_matches_to_return]
    return _return_close_values(id_to_keywords, search_term, max_matches_to_return, handle_typos)



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
#
# Copyright 2008-2021 Neongecko.com Inc. | All Rights Reserved
#
# Notice of License - Duplicating this Notice of License near the start of any file containing
# a derivative of this software is a condition of license for this software.
# Friendly Licensing:
# No charge, open source royalty free use of the Neon AI software source and object is offered for
# educational users, noncommercial enthusiasts, Public Benefit Corporations (and LLCs) and
# Social Purpose Corporations (and LLCs). Developers can contact developers@neon.ai
# For commercial licensing, distribution of derivative works or redistribution please contact licenses@neon.ai
# Distributed on an "AS IS” basis without warranties or conditions of any kind, either express or implied.
# Trademarks of Neongecko: Neon AI(TM), Neon Assist (TM), Neon Communicator(TM), Klat(TM)
# Authors: Guy Daniels, Daniel McKnight, Regina Bloomstine, Elon Gasper, Richard Leeds
#
# Specialized conversational reconveyance options from Conversation Processing Intelligence Corp.
# US Patents 2008-2021: US7424516, US20140161250, US20140177813, US8638908, US8068604, US8553852, US10530923, US10530924
# China Patent: CN102017585  -  Europe Patent: EU2156652  -  Patents Pending

import requests
from bs4 import BeautifulSoup
from abc import ABC
from html.parser import HTMLParser

from requests.exceptions import ConnectTimeout

from neon_utils.logger import LOG


def strip_tags(html):  # TODO: Document this! DM
    class MLStripper(HTMLParser, ABC):
        def __init__(self):
            super().__init__()
            self.reset()
            self.strict = False
            self.convert_charrefs = True
            self.fed = []

        def handle_data(self, d):
            self.fed.append(d)

        def get_data(self):
            return ''.join(self.fed)

    s = MLStripper()
    s.feed(html)
    return s.get_data()


def chunks(l, n):  # TODO: Document this! DM
    return [l[i:i + n] for i in range(0, len(l), n)]


def scrape_page_for_links(url: str) -> dict:
    """
    Scrapes the passed url for any links and returns a dictionary of link labels to URLs
    :param url: Web page to scrape
    :return: Lowercase names to links on page
    """
    import unicodedata
    available_links = {}
    retry_count = 0

    def _get_links(url):
        LOG.debug(url)
        if not str(url).startswith("http"):
            url = f"http://{url}"
        LOG.debug(url)
        html = requests.get(url, timeout=2.0).text
        soup = BeautifulSoup(html, 'lxml')
        # LOG.debug(html)
        # LOG.debug(soup)

        # Look through the page and find all anchor tags
        for i in soup.find_all("a", href=True):
            # LOG.debug(f"DM: found link: {i.text.rstrip()}")
            # LOG.debug(f"DM: found href: {i['href']}")

            if '://' not in i['href']:
                # Assume this is a relative address
                href = url + i['href'].lower()
            elif url.split('://')[1] in i['href']:
                href = i['href'].lower()
            else:
                href = None

            if href:
                available_links[unicodedata.normalize('NFKD', i.text.rstrip()
                                                      .replace(u'\u2013', '')
                                                      .replace(u'\u201d', '')
                                                      .replace(u'\u201c', '')
                                                      .replace('"', "")
                                                      .replace("'", "")
                                                      .replace("&apos;", "")
                                                      .lower())] = href
                LOG.debug("found link: " + unicodedata.normalize("NFKD", i.text.rstrip().replace(u"\u2013", "")
                                                                 .replace(u"\u201d", "").replace(u"\u201c", "")
                                                                 .replace('"', "").replace("'", "")
                                                                 .replace("&apos;", "").replace("\n", "").lower()))
                LOG.debug("found href: " + href)

        LOG.debug(available_links)

    try:
        _get_links(url)
    except ConnectTimeout:
        retry_count += 1
        if retry_count < 8:
            _get_links(url)
        else:
            raise ConnectTimeout
    except Exception as x:
        LOG.error(x)
        LOG.debug(available_links)
        raise ReferenceError
    return available_links
